@@ -5,7 +5,8 @@ from typing import Any, Literal, Dict, Sequence, Union, List
 from aiohttp.client import ClientSession
 from tinydb import TinyDB, Query
 from rich import print
-from centralcli import log, utils, config, CentralApi, cleaner, constants
+from centralcli import log, utils, config, CentralApi, cleaner, constants, Response
+from pathlib import Path
 
 import asyncio
 import time
@@ -16,13 +17,13 @@ try:
 except Exception:
     pass
 
-# try:
-#     import better_exceptions # noqa
-# except Exception:
-#     pass
-from rich.traceback import install
-install(show_locals=True)
-
+# TODO remove after TESTING NEW string matching lookup
+try:
+    from fuzzywuzzy import process # type: ignore noqa
+    FUZZ = True
+except Exception:
+    FUZZ = False
+    pass
 
 TinyDB.default_table_name = "devices"
 
@@ -124,8 +125,11 @@ class CentralObject:
                 self.serial,
                 self.mac,
                 self.ip,
-                f"s:{self.site}",
+                self.site,
             ]
+            parts = utils.strip_none(parts, strip_empty_obj=True)
+            if self.site:
+                parts[-1] = f"s:{parts[-1]}"
 
         return "|".join(
             [
@@ -133,6 +137,62 @@ class CentralObject:
             ]
         )
 
+# TODO Not used yet refactor to make consistent Response object available for when using contents of cache to avoid API call
+class CacheResponses:
+    def __init__(
+        self,
+        dev: Response = None,
+        site: Response = None,
+        template: Response = None,
+        group: Response = None
+    ) -> None:
+        self._dev = dev
+        self._site = site
+        self._template = template
+        self._group = group
+
+    def update_rl(self, resp: Response) -> Response:
+        """Returns provided Response object with the RateLimit info from the most recent API call.
+        """
+        _last_rl = sorted([r.rl for r in [self._dev, self._site, self._template, self._group] if r is not None], key=lambda k: k.remain_day)
+        if _last_rl:
+            resp.rl = _last_rl[0]
+        return resp
+
+    @property
+    def dev(self):
+        return self.update_rl(self._dev)
+
+    @dev.setter
+    def dev(self, resp: Response):
+        self._dev = resp
+
+    @property
+    def site(self):
+        return self.update_rl(self._site)
+
+    @site.setter
+    def site(self, resp: Response):
+        self._site = resp
+
+    @property
+    def template(self):
+        return self.update_rl(self._template)
+
+    @template.setter
+    def template(self, resp: Response):
+        self._template = resp
+
+    @property
+    def group(self):
+        return self.update_rl(self._group)
+
+    @group.setter
+    def group(self, resp: Response):
+        self._group = resp
+
+
+
 
 class Cache:
     def __init__(
@@ -146,6 +206,7 @@ class Cache:
         self.rl: str = ""  # TODO temp might refactor cache updates to return resp
         self.updated: list = []  # TODO change from list of methods to something easier
         self.central = central
+        self.responses = CacheResponses()
         if config.valid and config.cache_dir.exists():
             self.DevDB = TinyDB(config.cache_file)
             self.SiteDB = self.DevDB.table("sites")
@@ -156,6 +217,8 @@ class Cache:
             # it is updated anytime show logs is ran.
             self.LogDB = self.DevDB.table("logs")
             self.EventDB = self.DevDB.table("events")
+            self.HookConfigDB = self.DevDB.table("wh_config")
+            self.HookDataDB = self.DevDB.table("wh_data")
             self._tables = [self.DevDB, self.SiteDB, self.GroupDB, self.TemplateDB]
             self.Q = Query()
             if data:
@@ -191,6 +254,10 @@ class Cache:
     def events(self) -> list:
         return self.EventDB.all()
 
+    @property
+    def event_ids(self) -> list:
+        return [f'{x["id"]}' for x in self.EventDB.all()]
+
     @property
     def group_names(self) -> list:
         return [g["name"] for g in self.GroupDB.all()]
@@ -199,41 +266,77 @@ class Cache:
     def templates(self) -> list:
         return self.TemplateDB.all()
 
+    @property
+    def hook_config(self) -> list:
+        return self.HookConfigDB.all()
+
+    @property
+    def hook_data(self) -> list:
+        return self.HookDataDB.all()
+
+    @property
+    def hook_active(self) -> list:
+        return [h for h in self.HookDataDB.all() if h["state"].lower() == "open"]
+
     @property
     def all(self) -> dict:
         return {t.name: getattr(self, t.name) for t in self._tables}
 
+    async def get_hooks_by_serial(self, serial):
+        return self.HookDataDB.get(self.Q.device_id == serial)
+
     @staticmethod
     def account_completion(incomplete: str,):
         for a in config.defined_accounts:
             if a.lower().startswith(incomplete.lower()):
                 yield a
 
-    def smg_kw_completion(self, incomplete: str, args: List[str] = []):
+    # TODO maybe build script to gather completion and help text and place in flat file
+    def method_test_completion(self, incomplete: str, args: List[str] = []):
+        if self.central is not None:
+            methods = [
+                d for d in self.central.__dir__()
+                if not d.startswith("__")
+            ]
+
+        import importlib
+        bpdir = Path(__file__).parent / "boilerplate"
+        all_calls = [
+            importlib.import_module(f"centralcli.{bpdir.name}.{f.stem}") for f in bpdir.iterdir()
+            if not f.name.startswith("_") and f.suffix == ".py"
+        ]
+        for m in all_calls:
+            methods += [
+                d for d in m.AllCalls().__dir__()
+                if not d.startswith("__")
+            ]
+
+        for m in sorted(methods):
+            if m.startswith(incomplete):
+                yield m
+
+
+
+    def smg_kw_completion(self, ctx: typer.Context, incomplete: str, args: List[str] = []):
         kwds = ["group", "mac", "serial"]
         out = []
+        args = [v for k, v in ctx.params.items() if v and k[:2] in ["kw", "va"]]
         if args[-1].lower() == "group":
             out = [m for m in self.group_completion(incomplete, args)]
             for m in out:
                 yield m
         elif args[-1].lower() == "serial":
-            # out = [m for m in self.serial_pfx_completion(incomplete, args)]
             out = ["|", "<SERIAL NUMBER>"]
             if incomplete:
                 out.append(incomplete)
             for m in out:
                 yield m
         elif args[-1].lower() == "mac":
-            # out = [m for m in self.mac_oui_completion(incomplete, args)]
             out = ["|", "<MAC ADDRESS>"]
             for m in out:
                 yield m
 
         else:
-            # print(args[-2], incomplete)
-            # if args[-2].lower() == "mac":
-            #     # TODO not sure why but colons in prev arg breaks completion for next kw
-            #     args[-1] = args[-1].strip(":-.")
             for kw in kwds:
                 if kw not in args and kw.lower().startswith(incomplete):
                     yield kw
@@ -250,7 +353,6 @@ class Cache:
         args: List[str] = None,
     ):
         dev_type = None
-
         if args:
             if args[-1].lower() in ["gateways", "clients", "server"]:
                 dev_type = "gw"
@@ -265,7 +367,10 @@ class Cache:
             completion=True,
         )
         out = []
+        args = args or []
         if match:
+            # remove devices that are already on the command line
+            match = [m for m in match if m.name not in args]
             for m in sorted(match, key=lambda i: i.name):
                 if m.name.startswith(incomplete):
                     out += [tuple([m.name, m.help_text])]
@@ -284,6 +389,7 @@ class Cache:
 
     def dev_kwarg_completion(
         self,
+        ctx: typer.Context,
         incomplete: str,
         args: List[str] = None,
     ):
@@ -292,23 +398,30 @@ class Cache:
         i.e. cencli move dev1 dev2 dev3 site site_name group group_name
 
         Args:
+            ctx (typer.Context): Provided automatically by typer
             incomplete (str): The incomplete word for autocompletion
             args (List[str], optional): The prev args passed into the command.
+            # TODO verify and remove
+                Ignore this param params are pulled from ctx.params provided by typer
 
         Yields:
             tuple: matching completion string, help text
         """
-        if args[-1].lower() == "group":
+        args = [k for k, v in ctx.params.items() if v and k[:2] not in ["kw", "va"]]
+        args += [v for k, v in ctx.params.items() if v and k[:2] in ["kw", "va"]]
+        p = ctx.params
+        a = ctx.__dict__
+        if args and args[-1].lower() == "group":
             out = [m for m in self.group_completion(incomplete, args)]
             for m in out:
                 yield m
 
-        elif args[-1].lower() == "site":
+        elif args and args[-1].lower() == "site":
             out = [m for m in self.site_completion(incomplete, args)]
             for m in out:
                 yield m
 
-        elif args[-1].lower() == "ap":
+        elif args and args[-1].lower() == "ap":
             out = [m for m in self.dev_completion(incomplete, args)]
             for m in out:
                 yield m
@@ -317,9 +430,11 @@ class Cache:
             out = []
             if len(args) > 1:
                 if "site" not in args and "site".startswith(incomplete.lower()):
-                    out += ("site", )
+                    _help = "move device(s) to a different site" if ctx.info_name == "move" else f"{ctx.info_name} ... site"  # TODO the fallback ... need to check which commands use this.
+                    out += [("site", _help)]
                 if "group" not in args and "group".startswith(incomplete.lower()):
-                    out += ("group", )
+                    _help = "move device(s) to a different group" if ctx.info_name == "move" else f"{ctx.info_name} ... group"  # TODO the fallback ... need to check which commands use this.
+                    out += [("group", _help)]
 
             if "site" not in args and "group" not in args:
                 out += [m for m in self.dev_completion(incomplete, args)]
@@ -328,7 +443,160 @@ class Cache:
                 out += ["|", "<cr>"]
 
             for m in out:
-                yield m
+                yield m if isinstance(m, tuple) else (m, f"{ctx.info_name} ... {m}")
+
+    def dev_switch_ap_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Device completion for returning matches that are either switch or AP
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+        """
+        match = self.get_dev_identifier(incomplete, dev_type=["switch", "ap"], completion=True)
+
+        out = []
+        if match:
+            for m in sorted(match, key=lambda i: i.name):
+                out += [tuple([m.name, m.help_text])]
+
+        for m in out:
+            yield m[0], m[1]
+
+    def dev_ap_gw_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Completion for argument that can be either group or device.
+
+        use lambda function to pass dev_type
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+        """
+        dev_types = ["ap", "gw"]
+        match = [m for m in self.dev_completion(incomplete) if m.generic_type in dev_types]
+
+        out = []
+        if match:
+            for m in sorted(match, key=lambda i: i.name):
+                out += [tuple([m.name, m.help_text])]
+
+        for m in out:
+            yield m[0], m[1]
+
+    def dev_gw_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Completion for device idens where only gateways are valid.
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+
+        Yields:
+            tuple: name and help_text for the device
+        """
+        # match = [m for m in self.dev_completion(incomplete) if m.generic_type == "gw"]
+        match = self.get_identifier(incomplete, ["dev"], device_type="gw", completion=True)
+
+        out = []
+        if match:
+            for m in sorted(match, key=lambda i: i.name):
+                out += [tuple([m.name, m.help_text])]
+
+        for m in out:
+            yield m[0], m[1]
+
+    def group_dev_ap_gw_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Completion for argument that can be either group or device.
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+        """
+        dev_types = ["ap", "gw"]
+        dev_match = self.get_dev_identifier(incomplete, dev_type=dev_types, completion=True)
+        match = [*self.get_group_identifier(incomplete, completion=True), *dev_match]
+
+        out = []
+        if match:
+            for m in sorted(match, key=lambda i: i.name):
+                out += [tuple([m.name, m.help_text])]
+
+        for m in out:
+            yield m[0], m[1]
+
+    def group_dev_gw_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Completion for argument that can be either group or a gateway.
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+        """
+        # dev_types = ["gw"]
+        # dev_match = self.get_dev_identifier(incomplete, dev_type=dev_types, completion=True)
+        # match = [*self.get_group_identifier(incomplete, completion=True), *dev_match]
+        match = self.get_identifier(incomplete, ["group", "dev"], device_type="gw", completion=True)
+
+        out = []
+        if match:
+            for m in sorted(match, key=lambda i: i.name):
+                out += [tuple([m.name, m.help_text])]
+
+        for m in out:
+            yield m[0], m[1]
+
+    # FIXME completion doesn't pop args need ctx: typer.Context and reference ctx.params which is dict?
+    def send_cmds_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        """Completion for argument that can be either group, site, or a gateway or keyword "commands".
+
+        Args:
+            incomplete (str): The last partial or full command before completion invoked.
+            args (List[str], optional): The previous arguments/commands on CLI. Defaults to None.
+        """
+        if args[-1] == "all":
+            yield "commands"
+        elif args[-1] in ["commands", "file"]:
+            yield None
+        elif args[-1] not in ["group", "site", "devices"]:
+            yield "commands"
+        else:
+            if args[-1] == "group":
+                db = "group"
+            elif args[-1] == "site":
+                db = "site"
+            else:
+                db = "dev"
+
+            match = self.get_identifier(incomplete, [db], device_type="gw", completion=True)
+
+            out = []
+            if match:
+                for m in sorted(match, key=lambda i: i.name):
+                    out += [tuple([m.name if " " not in m.name else f"'{m.name}'", m.help_text])]
+
+                for m in out:
+                    yield m[0], m[1]
 
     def group_completion(
         self,
@@ -342,11 +610,27 @@ class Cache:
         out = []
         if match:
             for m in sorted(match, key=lambda i: i.name):
-                out += [tuple([m.name, m.help_text])]
+                if m.name not in args:
+                    out += [tuple([m.name, m.help_text])]
 
         for m in out:
             yield m[0], m[1]
 
+    def event_completion(
+        self,
+        incomplete: str,
+        args: List[str] = None,
+    ):
+        for event in self.events:
+            if event["id"].startswith(incomplete):
+                yield event["id"], f"{event['id']}|{event['device'].split('Group:')[0].rstrip()}"
+
+        # for match, help_txt in sorted(matches, key=lambda i: int(i[0])):
+        #     yield match, help_txt
+        # for event_id in self.event_ids:
+        #     if event_id.startswith(incomplete):
+        #         yield event_id, f"Details for Event with id {event_id}"
+
     # TODO add support for zip code city state etc.
     def site_completion(
         self,
@@ -359,6 +643,7 @@ class Cache:
         )
         out = []
         if match:
+            match = [m for m in match if m.name not in args]
             for m in sorted(match, key=lambda i: i.name):
                 out += [tuple([m.name if " " not in m.name else f"'{m.name}'", m.help_text])]
 
@@ -376,6 +661,7 @@ class Cache:
         )
         out = []
         if match:
+            match = [m for m in match if m.name not in args]
             for m in sorted(match, key=lambda i: i.name):
                 out += [tuple([m.name, m.help_text])]
 
@@ -523,6 +809,7 @@ class Cache:
             self.rl = str(resp.rl)
             resp.output = utils.listify(resp.output)
             resp.output = cleaner.get_devices(resp.output)
+            self.responses.dev = resp
             # resp.output = [
             #     {
             #         k: v if k != "type" else get_cencli_devtype(v) for k, v in r.items()
@@ -541,13 +828,14 @@ class Cache:
         if data:
             data = utils.listify(data)
             if not remove:
+                data = [{k.replace("site_", ""): v for k, v in d.items()} for d in data]
                 return self.SiteDB.insert_multiple(data)
             else:
                 doc_ids = []
                 for qry in data:
                     # provided list of site_ids to remove
                     if isinstance(qry, (int, str)) and str(qry).isdigit():
-                        doc_ids += [self.SiteDB.get((self.Q.id == int(qry))).doc_id]
+                        doc_ids += [self.SiteDB.get((self.Q.id == qry)).doc_id]
                     else:
                         # list of dicts with {search_key: value_to_search_for}
                         if len(qry.keys()) > 1:
@@ -559,13 +847,13 @@ class Cache:
             resp = await self.central.get_all_sites()
             if resp.ok:
                 resp.output = utils.listify(resp.output)
+                resp.output = [{k.replace("site_", ""): v for k, v in d.items()} for d in resp.output]
+                self.responses.site = resp
                 # TODO time this to see which is more efficient
-                # start = time.time()
                 # upd = [self.SiteDB.upsert(site, cond=self.Q.id == site.get("id")) for site in site_resp.output]
                 # upd = [item for in_list in upd for item in in_list]
                 self.updated.append(self.central.get_all_sites)
                 self.SiteDB.truncate()
-                # print(f" site db Done: {time.time() - start}")
                 update_res = self.SiteDB.insert_multiple(resp.output)
                 if False in update_res:
                     log.error("Tiny DB returned an error during site db update")
@@ -587,7 +875,9 @@ class Cache:
         else:
             resp = await self.central.get_all_groups()
             if resp.ok:
+                resp.output = cleaner.get_all_groups(resp.output)
                 resp.output = utils.listify(resp.output)
+                self.responses.group = resp
                 self.updated.append(self.central.get_all_groups)
                 self.GroupDB.truncate()
                 update_res = self.GroupDB.insert_multiple(resp.output)
@@ -596,11 +886,18 @@ class Cache:
             return resp
 
     async def update_template_db(self):
-        groups = self.groups if self.central.get_all_groups in self.updated else None
+        # groups = self.groups if self.central.get_all_groups in self.updated else None
+        if self.central.get_all_groups not in self.updated:
+            gr_resp = await self.update_group_db()
+            if not gr_resp.ok:
+                return gr_resp
+
+        groups = self.groups
         resp = await self.central.get_all_templates(groups=groups)
         if resp.ok:
             resp.output = utils.listify(resp.output)
             self.updated.append(self.central.get_all_templates)
+            self.responses.template = resp
             self.TemplateDB.truncate()
             update_res = self.TemplateDB.insert_multiple(resp.output)
             if False in update_res:
@@ -615,6 +912,39 @@ class Cache:
         self.EventDB.truncate()
         return self.EventDB.insert_multiple(log_data)
 
+    def update_hook_config_db(self, data: List[Dict[str, Any]], remove: bool = False) -> bool:
+        data = utils.listify(data)
+        self.HookConfigDB.truncate()
+        return self.HookConfigDB.insert_multiple(data)
+
+    async def update_hook_data_db(self, data: List[Dict[str, Any]]) -> bool:
+        data = utils.listify(data)
+        rem_data = []
+        add_data = []
+        for d in data:
+            if d.get("state", "") == "Close":
+                match = self.HookDataDB.get((self.Q.id == d["id"]))
+                if match is not None:
+                    rem_data += [match.doc_id]
+            else:
+                add_data += [d]
+
+        if rem_data and add_data:
+            log.error("update_hook_data_db called with both open and closed notifications")
+
+        if rem_data:
+            log.info(f"Removing {rem_data} from HookDataDB")
+            return self.HookDataDB.remove(doc_ids=rem_data)
+        elif add_data:
+            data = [*self.hook_active, *add_data]
+            self.HookDataDB.truncate()
+            return self.HookDataDB.insert_multiple(data)
+        else:
+            data = [*self.hook_active, *data]
+            self.HookDataDB.truncate()
+            return self.HookDataDB.insert_multiple(data)
+
+    # TODO cache.groups cache.devices etc change to Response object with data in output.  So they can be leveraged in commands with all atributes
     async def _check_fresh(self, dev_db: bool = False, site_db: bool = False, template_db: bool = False, group_db: bool = False):
         update_funcs, db_res = [], []
         if dev_db:
@@ -654,18 +984,20 @@ class Cache:
             refresh = True
 
         if refresh or not config.cache_file.is_file() or not config.cache_file.stat().st_size > 0:
+            _word = "Refreshing" if refresh else "Populating"
+            print(f"[cyan]-- {_word} Identifier mapping Cache --[/cyan]", end="")
+
             start = time.time()
-            print("[cyan]-- Refreshing Identifier mapping Cache --[/cyan]", end="")
             db_res = asyncio.run(self._check_fresh(dev_db=dev_db, site_db=site_db, template_db=template_db, group_db=group_db))
+
             if not all([r.ok for r in db_res]):
-                # if db_res and False in db_res:
                 res_map = ["dev_db", "site_db", "template_db", "group_db"]
-                res_map = ", ".join([db for idx, db in enumerate(res_map) if not db_res(idx)])
+                res_map = ", ".join([db for idx, db in enumerate(res_map[0:len(db_res)]) if not db_res[idx]])
                 # log.error(f"TinyDB returned error ({res_map}) during db update")
                 self.central.spinner.fail(f"Cache Refresh Returned an error updating ({res_map})")
             else:
                 self.central.spinner.succeed(f"Cache Refresh Completed in {round(time.time() - start, 2)} sec")
-            log.info(f"Cache Refreshed in {round(time.time() - start, 2)} seconds")
+            log.info(f"Cache Refreshed in {round(time.time() - start, 2)} seconds", show=False)
             # typer.secho(f"-- Cache Refresh Completed in {round(time.time() - start, 2)} sec --", fg="cyan")
 
     def handle_multi_match(
@@ -714,21 +1046,23 @@ class Cache:
         self,
         qry_str: str,
         qry_funcs: Sequence[str],
-        device_type: str = None,
+        device_type: Union[str, List[str]] = None,
         group: str = None,
         multi_ok: bool = False,
+        all: bool = False,
         completion: bool = False,
-    ) -> CentralObject:
+    ) -> Union[CentralObject, List[CentralObject]]:
         """Get Identifier when iden type could be one of multiple types.  i.e. device or group
 
         Args:
             qry_str (str): The query string provided by user.
             qry_funcs (Sequence[str]): Sequence of strings "dev", "group", "site", "template"
-            device_type (str, optional): str indicating what devices types are valid for dev idens.
+            device_type (Union[str, List[str]], optional): Restrict matches to specific dev type(s).
                 Defaults to None.
             group (str, optional): applies to get_template_identifier, Only match if template is in this group.
                 Defaults to None.
             multi_ok (bool, optional): DEPRECATED, NO LONGER USED
+            all (bool, optional): For use in completion, adds keyword "all" to valid completion.
             completion (bool, optional): If function is being called for AutoCompletion purposes. Defaults to False.
                 When called for completion it will fail silently and will return multiple when multiple matches are found.
 
@@ -736,10 +1070,13 @@ class Cache:
             typer.Exit: If not ran for completion, and there is no match, exit with code 1.
 
         Returns:
-            CentralObject
+            CentralObject or list[CentralObject, ...]
         """
-        # TODO remove multi_ok once verified refs are removed
+        # DEPRECATED remove multi_ok once verified refs are removed
+        if multi_ok:
+            log.warning("DEV NOTE: get_identifier called with deprecated kwarg mutli_ok", show=True)
         match = None
+        device_type = utils.listify(device_type)
         default_kwargs = {"retry": False, "completion": completion, "silent": True}
         for _ in range(0, 2):
             for q in qry_funcs:
@@ -763,6 +1100,10 @@ class Cache:
                 )
 
         if completion:
+            if all:
+                if "all".startswith(qry_str.lower()):
+                    match = utils.listify(match)
+                    match += CentralObject("dev", {"name": "all", "help_text": "All Devices"})
             return match
 
         if not match:
@@ -772,7 +1113,7 @@ class Cache:
     def get_dev_identifier(
         self,
         query_str: Union[str, List[str], tuple],
-        dev_type: str = None,
+        dev_type: Union[str, List[str]] = None,
         # ret_field: str = "serial",       # TODO ret_field believe to be deprecated, now returns an object with all attributes
         retry: bool = True,
         # multi_ok: bool = True,          # TODO multi_ok also believe to be deprecated check
@@ -792,7 +1133,7 @@ class Cache:
             # Try exact match
             match = self.DevDB.search(
                 (self.Q.name == query_str)
-                | (self.Q.ip.test(lambda v: v.split("/")[0] == query_str))
+                | (self.Q.ip.test(lambda v: v and v.split("/")[0] == query_str))
                 | (self.Q.mac == utils.Mac(query_str).cols)
                 | (self.Q.serial == query_str)
             )
@@ -833,24 +1174,29 @@ class Cache:
 
         all_match = None
         if dev_type:
-            all_match = match
-            match = [d for d in match if d.generic_type.lower() in "".join(dev_type[0:len(d.generic_type)]).lower()]
+            all_match = match.copy()
+            dev_type = utils.listify(dev_type)
+            match = []
+            for _dev_type in dev_type:
+                match += [d for d in all_match if d.generic_type.lower() in "".join(_dev_type[0:len(d.generic_type)]).lower()]
+
+        if completion:
+            return match or []
 
         if match:
-            if completion:
-                return match
-
-            elif len(match) > 1:
+            if len(match) > 1:
                 match = self.handle_multi_match(match, query_str=query_str,)  # multi_ok=multi_ok)
 
             return match[0]
+
         elif retry:
             log.error(f"Unable to gather device info from provided identifier {query_str}", show=not silent)
             if all_match:
                 # all_match = all_match[-1]
                 all_match_msg = f"{', '.join(m.name for m in all_match[0:5])}{', ...' if len(all_match) > 5 else ''}"
+                _dev_type_str = ", ".join(dev_type)
                 log.error(
-                    f"The Following devices matched {all_match_msg} excluded as device type != {dev_type}",
+                    f"The Following devices matched {all_match_msg} excluded as device type != [{_dev_type_str}]",
                     show=True,
                 )
             raise typer.Exit(1)
@@ -870,6 +1216,10 @@ class Cache:
         if isinstance(query_str, (list, tuple)):
             query_str = " ".join(query_str)
 
+
+        if completion and query_str == "":
+            return [CentralObject("site", s) for s in self.sites]
+
         match = None
         for _ in range(0, 2 if retry else 1):
             # try exact site match
@@ -880,9 +1230,11 @@ class Cache:
                 | (self.Q.address == query_str)
                 | (self.Q.city == query_str)
                 | (self.Q.state == query_str)
-                | (self.Q.state.test(lambda v: constants.state_abbrev_to_pretty.get(query_str.upper(), "").title() == v.title()))
+                | (self.Q.state.test(lambda v: constants.state_abbrev_to_pretty.get(query_str.upper(), query_str).title() == v.title()))
             )
 
+            # raise ValueError(f'>{query_str}<, {type(query_str)}, {", ".join([m["name"] for m in match])} {bool(match)}')
+
             # retry with case insensitive name & address match if no match with original query
             if not match:
                 match = self.SiteDB.search(
@@ -913,20 +1265,20 @@ class Cache:
                 self.check_fresh(refresh=True, site_db=True)
             if match:
                 match = [CentralObject("site", s) for s in match]
+                # raise ValueError(f'>{query_str}<, {type(query_str)}, {", ".join([m.name for m in match])}')
                 break
 
-        if match:
-            if completion:
-                return match
+        if completion:
+            return match
 
+        if match:
             if len(match) > 1:
-                match = self.handle_multi_match(match, query_str=query_str, query_type="site", multi_ok=multi_ok)
+                match = self.handle_multi_match(match, query_str=query_str, query_type="site",)  # multi_ok=multi_ok)
 
             return match[0]
 
-        # FIXME check this ... would think if not retry also doesn't match other get_*_identifier func logic
         elif retry:
-            log.error(f"Unable to gather site {ret_field} from provided identifier {query_str}", show=not silent)
+            log.error(f"Unable to gather site info from provided identifier {query_str}", show=not silent)
             raise typer.Exit(1)
 
     def get_group_identifier(
@@ -985,19 +1337,25 @@ class Cache:
                 )
 
             if not match and retry and self.central.get_all_groups not in self.updated:
-                typer.secho(f"No Match Found for {query_str}, Updating group Cache", fg="red")
-                self.check_fresh(refresh=True, group_db=True)
+                print(f"[bright_red]No Match found for[/] [cyan]{query_str}[/].")
+                if FUZZ:
+                    fuzz_match, fuzz_confidence = process.extract(query_str, [g["name"] for g in self.groups], limit=1)[0]
+                    if fuzz_confidence >= 70 and typer.confirm(f"Did you mean {fuzz_match}?"):
+                        match = self.GroupDB.search(self.Q.name == fuzz_match)
+                if not match:
+                    typer.secho(f"No Match Found for {query_str}, Updating group Cache", fg="red")
+                    self.check_fresh(refresh=True, group_db=True)
                 _ += 1
             if match:
                 match = [CentralObject("group", g) for g in match]
                 break
 
-        if match:
-            if completion:
-                return match
+        if completion:
+            return match or []
 
+        if match:
             if len(match) > 1:
-                match = self.handle_multi_match(match, query_str=query_str, query_type="group", multi_ok=multi_ok)
+                match = self.handle_multi_match(match, query_str=query_str, query_type="group",)  # multi_ok=multi_ok)
 
             return match[0]
 
@@ -1025,6 +1383,13 @@ class Cache:
     ) -> CentralObject:
         """Allows case insensitive template match by template name"""
         retry = False if completion else retry
+        if not query_str and completion:
+            return [CentralObject("template", data=t) for t in self.templates]
+
+        # TODO verify and remove
+        if multi_ok:
+            log.error("depricated parameter multi_ok sent to get_template_identifier.")
+
         match = None
         for _ in range(0, 2 if retry else 1):
             # exact
@@ -1069,7 +1434,7 @@ class Cache:
                     match,
                     query_str=query_str,
                     query_type="template",
-                    multi_ok=multi_ok,
+                    # multi_ok=multi_ok,
                 )
 
             return match[0]
@@ -1085,10 +1450,13 @@ class Cache:
         if "audit_trail" in query:
             return query
         elif query == "":  # tab completion
-            return [x["id"] for x in self.logs]
+            return ["cencli", *[x["id"] for x in self.logs]]
 
         try:
 
+            if "cencli".startswith(query.lower()):
+                return ["cencli"]
+
             match = self.LogDB.search(self.Q.id == int(query))
             if not match:
                 log.warning(f"Unable to gather log id from short index query {query}", show=True)
