@@ -4,23 +4,25 @@
 """
 Collection of functions used to clean output from Aruba Central API into a consistent structure.
 """
-from pathlib import Path
-import sys
 import functools
-import logging
-from typing import Dict, List, Any, Union
-import pendulum
 import ipaddress
+import logging
+import sys
+from pathlib import Path
+from typing import Any, Dict, List, Union
 
+import pendulum
+from rich.console import Console
+from rich.markup import escape
 
 # Detect if called from pypi installed package or via cloned github repo (development)
 try:
-    from centralcli import utils, constants
+    from centralcli import constants, utils
 except (ImportError, ModuleNotFoundError) as e:
     pkg_dir = Path(__file__).absolute().parent
     if pkg_dir.name == "centralcli":
         sys.path.insert(0, str(pkg_dir.parent))
-        from centralcli import utils, constants
+        from centralcli import constants, utils
     else:
         print(pkg_dir.parts)
         raise e
@@ -29,7 +31,9 @@ except (ImportError, ModuleNotFoundError) as e:
 def epoch_convert(func):
     @functools.wraps(func)
     def wrapper(epoch):
-        if len(str(int(epoch))) > 10:
+        # FIXME don't know why started getting ValueError: invalid literal for int() with base 10: '24 minutes 20 seconds'
+        # appears show all was hitting the cleaner 2x
+        if str(epoch).isdigit() and len(str(int(epoch))) > 10:
             epoch = epoch / 1000
         return func(epoch)
 
@@ -62,6 +66,7 @@ def _time_diff_words(epoch: float) -> str:
 @epoch_convert
 def _log_timestamp(epoch: float) -> str:
     return pendulum.from_timestamp(epoch, tz="local").format("MMM DD h:mm:ss A")
+    # DEBUG return f'{epoch} | {pendulum.from_timestamp(epoch, tz="local").format("MMM DD h:mm:ss A")}'
 
 
 def _short_connection(value: str) -> str:
@@ -76,7 +81,7 @@ def _serial_to_name(sernum: str) -> str:
     ):
         return sernum
 
-    match = cache.get_dev_identifier(sernum)
+    match = cache.get_dev_identifier(sernum, retry=False, silent=True)
     if not match:
         return sernum
 
@@ -84,8 +89,11 @@ def _serial_to_name(sernum: str) -> str:
 
 
 def _extract_names_from_id_name_dict(id_name: dict) -> str:
-    names = [x.get("name", "Error") for x in id_name]
-    return ", ".join(names)
+    if isinstance(id_name, dict) and "id" in id_name and "name" in id_name:
+        names = [x.get("name", "Error") for x in id_name]
+        return ", ".join(names)
+    else:
+        return id_name
 
 
 def _extract_event_details(details: List[dict]) -> dict:
@@ -107,13 +115,14 @@ _short_value = {
     "last_modified": _convert_epoch,
     "lease_start_ts": _log_timestamp,
     "lease_end_ts": _log_timestamp,
+    "acknowledged_timestamp": _log_timestamp,
     "lease_time": _duration_words,
     "lease_time_left": _duration_words,
     "ts": _log_timestamp,
     "timestamp": _log_timestamp,
     "Unknown": "?",
     "HPPC": "SW",
-    # "labels": _extract_names_from_id_name_dict,
+    "labels": _extract_names_from_id_name_dict,
     "sites": _extract_names_from_id_name_dict,
     "ACCESS POINT": "AP",
     "GATEWAY": "GW",
@@ -159,6 +168,9 @@ _short_key = {
     "vlan_id": "pvid",
     "free_ip_addr_percent": "free ip %",
     "events_details": "details",
+    # "acknowledged": "ack",
+    "acknowledged_by": "ack by",
+    "acknowledged_timestamp": "ack time",
 }
 
 
@@ -219,6 +231,9 @@ def short_key(key: str) -> str:
 
 def short_value(key: str, value: Any):
     # _unlist(value)
+    # Run any inner dicts through cleaner funcs
+    if isinstance(value, dict):
+        value = {short_key(k): v if k not in _short_value else _short_value[k](v) for k, v in value.items()}
 
     if isinstance(value, (str, int, float)):
         return (
@@ -234,13 +249,22 @@ def short_value(key: str, value: Any):
     return short_key(key), _unlist(value)
 
 
-def _get_group_names(
-    data: List[
-        str,
-    ]
-) -> list:
+def _get_group_names(data: List[List[str],]) -> list:
+    """Convert list of single item lists to a list of strs
+
+    Also removes "unprovisioned" group as it has no value for our
+    purposes, and moves default group to beginning of list.
+
+    Args:
+        data (List[List[str],]): Central response payload
+
+    Returns:
+        list: List of strings with "unprovisioned" group removed
+        and default moved to front.
+    """
     groups = [g for _ in data for g in _ if g != "unprovisioned"]
-    groups.insert(0, groups.pop(groups.index("default")))
+    if "default" in groups:
+        groups.insert(0, groups.pop(groups.index("default")))
     return groups
 
 
@@ -258,7 +282,6 @@ def _client_concat_associated_dev(
     verbose: bool = False,
     cache=None,
 ) -> Dict[str, Any]:
-    # FIXME
     if cache is None:
         from centralcli import cache
     strip_keys = [
@@ -271,10 +294,10 @@ def _client_concat_associated_dev(
     ]
     dev, _gw, data["gateway"] = "", "", {}
     if data.get("associated_device"):
-        dev = cache.get_dev_identifier(data["associated_device"], ret_field="name")
+        dev = cache.get_dev_identifier(data["associated_device"],)
 
     if data.get("gateway_serial"):
-        _gw = cache.get_dev_identifier(data["gateway_serial"], ret_field="name")
+        _gw = cache.get_dev_identifier(data["gateway_serial"],)
         _gateway = {
             "name": _gw.name,
             "serial": data.get("gateway_serial", ""),
@@ -367,13 +390,12 @@ def get_clients(
 
 def strip_no_value(data: List[dict]) -> List[dict]:
     """strip out any columns that have no value in any row"""
+    no_val_strings = ["Unknown", "NA", "None", "--", ""]
     no_val: List[List[int]] = [
         [
             idx
             for idx, v in enumerate(id.values())
-            if not isinstance(v, bool)
-            and not v
-            or (isinstance(v, str) and v == "Unknown" or isinstance(v, str) and v == "NA" or isinstance(v, str) and v == "--")
+            if (not isinstance(v, bool) and not v) or (isinstance(v, str) and v and v in no_val_strings)
         ]
         for id in data
     ]
@@ -403,6 +425,8 @@ def sort_result_keys(data: List[dict], order: List[str] = None) -> List[dict]:
         to_front = [
             "vlan_id",
             "name",
+            "status",
+            "type",
             "model",
             'mode',
             "vlan_desc",
@@ -420,12 +444,12 @@ def sort_result_keys(data: List[dict], order: List[str] = None) -> List[dict]:
             "group_name",
             "group",
             "site",
+            "labels",
             "addr_mode",
             "admin_mode",
             "oper_mode",
             "untagged_ports",
             "tagged_ports",
-            "status",
             "is_management_vlan",
             "is_jumbo_enabled",
             "is_voice_enabled",
@@ -436,7 +460,9 @@ def sort_result_keys(data: List[dict], order: List[str] = None) -> List[dict]:
             'mem_total',
             'mem_free',
             'firmware_version',
-            'firmware_backup_version'
+            'version',
+            'firmware_backup_version',
+            'oper_state_reason',
         ]
     to_front = [i for i in to_front if i in all_keys]
     _ = [all_keys.insert(0, all_keys.pop(all_keys.index(tf))) for tf in to_front[::-1]]
@@ -467,6 +493,8 @@ def get_devices(data: Union[List[dict], dict], sort: str = None) -> Union[List[d
         ]
     )
 
+    data = utils.listify(data)
+    data = sorted(data, key=lambda i: (i.get("site") or "", i.get("type") or "", i.get("name") or ""))
     # if sort and data and sort in data[-1]:
     #     return sorted(data, key=sort)
     # else:
@@ -489,7 +517,7 @@ def get_audit_logs(data: List[dict], cache_update_func: callable = None) -> List
     data = [dict(short_value(k, d.get(k)) for k in field_order) for d in data]
     data = strip_no_value(data)
 
-    if len(data) > 1 and cache_update_func:
+    if len(data) > 0 and cache_update_func:
         idx, cache_list = 1, []
         for d in data:
             if d.get("has details") is True:
@@ -505,6 +533,38 @@ def get_audit_logs(data: List[dict], cache_update_func: callable = None) -> List
     return data
 
 
+def get_alerts(data: List[dict],) -> List[dict]:
+    if isinstance(data, list) and "No Alerts" in data:
+        return data
+
+    field_order = [
+        "timestamp",
+        "severity",
+        "type",
+        # "group_name",
+        # "device info",
+        "description",
+        # "labels",
+        "acknowledged",
+        # "acknowledged_by",
+        # "acknowledged_timestamp",
+        # "state",
+    ]
+    for d in data:
+        # d["details"] = d.get("details") or {}
+        # d["device info"] = f"{d['details'].get('hostname', '')}|" \
+        #     f"{d.get('device_id', '')}|Group: {d.get('group_name', '')}".lstrip("|")
+        d["severity"] = d.get("severity", "INFO")
+        d["acknowledged"] = f'{"by " if d.get("acknowledged_by") else ""}' \
+            f'{d.get("acknowledged_by")}{" @ " if d.get("acknowledged_by") else ""}' \
+            f'{"" if not d.get("acknowledged_timestamp") else _log_timestamp(d["acknowledged_timestamp"])}' \
+
+    data = [dict(short_value(k, d[k]) for k in field_order if k in d) for d in data]
+    data = strip_no_value(data)
+
+    return data
+
+
 def get_event_logs(data: List[dict], cache_update_func: callable = None) -> List[dict]:
     # TODO accept verbose option and provide additional details with verbose. (currently just bypasses cleaner and displays yaml)
 
@@ -535,6 +595,7 @@ def get_event_logs(data: List[dict], cache_update_func: callable = None) -> List
         # for key in ["device_type", "hostname", "device_serial", "device_mac"]:
         #     del d[key]
 
+    # Stash event_details in cache indexed starting @ most recent event
     if len(data) > 1 and cache_update_func:
         idx, cache_list = 1, []
         for d in data:
@@ -542,7 +603,7 @@ def get_event_logs(data: List[dict], cache_update_func: callable = None) -> List
                 _details = {
                     k: v for inner in d["events_details"] for k, v in inner.items()
                 }
-                cache_list += [{"id": idx, "details": _details}]
+                cache_list += [{"id": str(idx), "device": d["device info"], "details": _details}]
                 d["id"] = idx
                 idx += 1
             else:
@@ -722,3 +783,144 @@ def get_dhcp(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
     #     for d in data
     # ]
     return data
+
+def get_template_details_for_device(data: str) -> dict:
+    """Convert form-data response to dict
+
+    Args:
+        data (str): string data with summary and optionally running config
+
+    Returns:
+        dict: dict with summary(dict), running config(str) and
+        central side config(str).
+    """
+    import json
+
+    summary, running_config, central_config = None, None, None
+    split_line = data.split("\n")[0].rstrip()
+    data_parts = [
+        d.lstrip().splitlines() for d in data.split(split_line)
+        if d.lstrip().startswith("Content-Disposition")
+    ]
+    for part in data_parts:
+        if 'name="Summary"' in part[0]:
+            summary = json.loads("\n".join(part[2:]))
+            summary = {k.lower(): v for k, v in summary.items()}
+        elif 'name="Device_running_config"' in part[0]:
+            running_config = "\n".join(part[2:]).rstrip()
+        elif 'name="Device_central_side_config"' in part[0]:
+            central_config = "\n".join(part[2:]).rstrip()
+
+    return {
+        "summary": summary,
+        "running_config": running_config,
+        "central_config": central_config
+    }
+
+def parse_caas_response(data: Union[dict, List[dict]]) -> List[str]:
+    """Parses Response Object from caas API updates output attribute
+
+    """
+    data = utils.unlistify(data)
+    out = []
+    lines = f"[reset]{'-' * 22}"
+
+    if data.get("_global_result", {}).get("status", '') == 0:
+        global_res = "[bright_green]Success[/bright_green]"
+    else:
+        global_res = "[red]Failure[/red]"
+    out += [lines, f"Global Result: {global_res}", lines]
+
+    if data.get("cli_cmds_result"):
+        out += ["\n -- [cyan bold]Command Results[/cyan bold] --"]
+        for cmd_resp in data["cli_cmds_result"]:
+            for _c, _r in cmd_resp.items():
+                _r_code = _r.get("status")
+                if _r_code == 0:
+                    _r_pretty = "[bright_green]OK[/bright_green]"
+                elif _r_code == 2:
+                    _r_pretty = "[dark_orange3]WARNING[/dark_orange3]"
+                else:
+                    _r_pretty = f"[red]ERROR[/red]" if _r_code == 1 else f"[red]ERROR ({_r_code})[/red]"
+
+                out += [f" [{_r_pretty}] {_c}"]
+                cmd_status = _r.get('status_str')
+                if cmd_status:
+                    _r_txt = f"[italic]{escape(cmd_status)}[/italic]"
+                    out += [lines, _r_txt, lines]
+
+        console = Console(emoji=False, record=True)
+        with console.capture():
+            console.print("\n".join(out))
+
+        out = console.export_text(styles=True)
+
+    return out
+
+def get_all_webhooks(data: list) -> list:
+    # TODO this is the default handling syntax
+    data = [
+        dict(short_value(k, d[k]) for k in d) for d in strip_no_value(data)
+    ]
+
+    return data
+
+def get_branch_health(data: list, down: bool = False, wan_down: bool = False) -> list:
+    field_order = [
+                "name",
+                "potential_issue",
+                "device_up",
+                "device_down",
+                "connected_count",
+                "failed_count",
+                "wan_tunnels_down",
+                "wan_tunnels_no_issue",
+                "wan_uplinks_down",
+                "wan_uplinks_no_issue",
+                "branch_cpu_high",
+                "branch_device_status_up",
+                "branch_device_status_down",
+                "cape_state",
+                "cape_state_dscr",
+                "cape_url",
+                "device_high_ch_2_4ghz",
+                "device_high_ch_5ghz",
+                "device_high_cpu",
+                "device_high_mem",
+                "device_high_noise_2_4ghz",
+                "device_high_noise_5ghz",
+                "insight_hi",
+                "insight_lo",
+                "insight_mi",
+                "score",
+                "silverpeak_state",
+                "silverpeak_state_summary",
+                "silverpeak_url",
+                "user_conn_health_score",
+                "wired_cpu_high",
+                "wired_device_status_down",
+                "wired_device_status_up",
+                "wired_mem_high",
+                "wlan_cpu_high",
+                "wlan_device_status_down",
+                "wlan_device_status_up",
+                "branch_mem_high",
+                "wlan_mem_high",
+    ]
+    # if wan_down:
+    #     test = lambda d: d["wan_uplinks_down"] > 0 or d["wan_tunnels_down"] > 0
+    # elif down:
+    #     test = lambda d: d["device_down"] > 0
+    # else:
+    #     test = lambda d: True
+
+    # data = [
+    #     dict(short_value(k, d[k]) for k in field_order) for d in data if test(d)
+    # ]
+    data = [
+        dict(short_value(k, d[k]) for k in field_order) for d in data
+    ]
+    data = strip_no_value(data)
+
+
+    return data
