@@ -22,7 +22,7 @@ except (ImportError, ModuleNotFoundError) as e:
         print(pkg_dir.parts)
         raise e
 
-from centralcli.constants import GatewayRole, LicenseTypes
+from centralcli.constants import GatewayRole, LicenseTypes, state_abbrev_to_pretty
 
 app = typer.Typer()
 color = utils.color
@@ -48,6 +48,7 @@ class AddGroupArgs(str, Enum):
 
 
 # TODO update completion with mac oui, serial prefix
+# FIXME Not all flows work on 2.5.5  I think license may be broken
 @app.command(short_help="Add a Device to Aruba Central.")
 def device(
     kw1: AddGroupArgs = typer.Argument(..., hidden=True, metavar="",),
@@ -57,8 +58,12 @@ def device(
     kw3: str = typer.Argument(None, metavar="", hidden=True, autocompletion=cli.cache.smg_kw_completion),
     val3: str = typer.Argument(None, metavar="group [GROUP]", help="pre-assign device to group",
                                autocompletion=cli.cache.smg_kw_completion),
+    kw4: str = typer.Argument(None, metavar="", hidden=True, autocompletion=cli.cache.smg_kw_completion),
+    val4: str = typer.Argument(None, metavar="site [SITE]", help="Assign newly added device to site",
+                               autocompletion=cli.cache.smg_kw_completion),
     _: str = typer.Argument(None, metavar="", hidden=True, autocompletion=cli.cache.null_completion),
     _group: str = typer.Option(None, "--group", autocompletion=cli.cache.group_completion, hidden=True),
+    _site: str = typer.Option(None, "--site", autocompletion=cli.cache.site_completion, hidden=True),
     license: List[LicenseTypes] = typer.Option(None, "--license", help="Assign license subscription(s) to device"),
     yes: bool = typer.Option(False, "-Y", help="Bypass confirmation prompts - Assume Yes"),
     yes_: bool = typer.Option(False, "-y", hidden=True),
@@ -69,12 +74,13 @@ def device(
                                 help="The Aruba Central Account to use (must be defined in the config)",),
 ) -> None:
     yes = yes_ if yes_ else yes
-    kwd_vars = [kw1, kw2, kw3]
-    vals = [val1, val2, val3]
+    kwd_vars = [kw1, kw2, kw3, kw4]
+    vals = [val1, val2, val3, val4]
     kwargs = {
         "mac": None,
         "serial": None,
         "group": None,
+        "site": None,
         "license": license
     }
 
@@ -82,12 +88,13 @@ def device(
         print("DEVELOPER NOTE Null completion item has value... being ignored.")
     for name, value in zip(kwd_vars, vals):
         if name and name not in kwargs:
-            print(f"[bright_red][blink]Error[/bright_red]: {name} in invalid")
+            print(f"[bright_red][blink]Error[/bright_red]: {name} is invalid")
             raise typer.Exit(1)
         else:
             kwargs[name] = value
 
     kwargs["group"] = kwargs["group"] or _group
+    kwargs["site"] = kwargs["site"] or _site
 
     # Error if both serial and mac are not provided
     if not kwargs["mac"] or not kwargs["serial"]:
@@ -99,7 +106,13 @@ def device(
 
     _msg = [f"Add device: [bright_green]{kwargs['serial_num']}|{kwargs['mac_address']}[/bright_green]"]
     if "group" in kwargs and kwargs["group"]:
+        _group = cli.cache.get_group_identifier(kwargs["group"])
+        kwargs["group"] = _group.name
         _msg += [f"\n  Pre-Assign to Group: [bright_green]{kwargs['group']}[/bright_green]"]
+    if "site" in kwargs and kwargs["site"]:
+        _site = cli.cache.get_site_identifier(kwargs["site"])
+        kwargs["site"] = _site.id
+        _msg += [f"\n  Assign to Site: [bright_green]{_site.name}[/bright_green]"]
     if "license" in kwargs and kwargs["license"]:
         _lic_msg = [lic._value_ for lic in kwargs["license"]]
         _lic_msg = _lic_msg if len(kwargs["license"]) > 1 else _lic_msg[0]
@@ -182,17 +195,17 @@ def group(
     # -- // Error on combinations that are not allowed by API \\ --
     if not aos10 and microbranch:
         print(
-            f"[bright_red]Microbranch is only valid if group is configured as AOS10 group ({color('--aos10')})."
+            f":x: [bright_red]Microbranch is only valid if group is configured as AOS10 group ({color('--aos10')})."
         )
         raise typer.Exit(1)
     if (mon_only_sw or mon_only_cx) and wired_tg:
-        print("[bright_red]Error: Monitor only is not valid for template group.")
+        print(":x: [bright_red]Error: Monitor only is not valid for template group.")
         raise typer.Exit(1)
     if not [t for t in allowed_types if allowed_types in ["cx", "sw"]] and (mon_only_sw or mon_only_cx):
-        print("[bright_red]Error: Monitor only is not valid without '--sw' or '--cx' (Allowed Device Types)")
+        print(":x: [bright_red]Error: Monitor only is not valid without '--sw' or '--cx' (Allowed Device Types)")
         raise typer.Exit(1)
     if gw_role and gw_role == "wlan" and not aos10:
-        print("[bright_red]WLAN role for Gateways requires the group be configured as AOS10 via --aos10 option.")
+        print(":x: [bright_red]WLAN role for Gateways requires the group be configured as AOS10 via --aos10 option.")
         raise typer.Exit(1)
     if all([x is None for x in [ap, sw, cx, gw]]):
         print("[green]No Allowed devices provided. Allowing all device types.")
@@ -233,6 +246,8 @@ def group(
             asyncio.run(
                 cli.cache.update_group_db({'name': group, 'template group': {'Wired': wired_tg, 'Wireless': wlan_tg}})
             )
+        else:
+            raise typer.Exit(1)
 
 
 # TODO autocompletion
@@ -292,6 +307,86 @@ def wlan(
         raise typer.Abort()
 
 
+@app.command(short_help="Add a site.")
+def site(
+    site_name: str = typer.Argument(...),
+    address: str = typer.Argument(None, help="street address"),
+    city: str = typer.Argument(None,),
+    state: str = typer.Argument(
+        None,
+        autocompletion=lambda incomplete: [
+        s for s in [
+            *list(state_abbrev_to_pretty.keys()),
+            *list(state_abbrev_to_pretty.values())
+            ]
+            if s.lower().startswith(incomplete.lower())
+        ]
+    ),
+    zipcode: int = typer.Argument(None,),
+    country: str = typer.Argument(None,),
+    lat: str = typer.Option(None,),
+    lon: str = typer.Option(None,),
+    yes: bool = typer.Option(False, "-Y", help="Bypass confirmation prompts - Assume Yes"),
+    yes_: bool = typer.Option(False, "-y", hidden=True),
+    default: bool = typer.Option(
+        False, "-d", is_flag=True, help="Use default central account", show_default=False
+    ),
+    debug: bool = typer.Option(
+        False, "--debug", envvar="ARUBACLI_DEBUG", help="Enable Additional Debug Logging",
+    ),
+    account: str = typer.Option(
+        "central_info",
+        envvar="ARUBACLI_ACCOUNT",
+        help="The Aruba Central Account to use (must be defined in the config)",
+    ),
+) -> None:
+    """Perform batch Add operations using import data from file."""
+    yes = yes_ if yes_ else yes
+    kwargs = {
+        "address": address,
+        "city": city,
+        "state": state,
+        "zipcode": str(zipcode),
+        "country": country,
+        "latitude": lat,
+        "longitude": lon
+    }
+    address_fields = {k: v for k, v in kwargs.items() if v}
+
+    print(f"Add Site: [cyan]{site_name}[reset]:")
+    _ = [print(f"  {k}: {v}") for k, v in address_fields.items()]
+    if yes or typer.confirm(f"\nProceed?", abort=True):
+        resp = cli.central.request(cli.central.create_site, site_name, **address_fields)
+        cli.display_results(resp)
+        if resp:
+            asyncio.run(cli.cache.update_site_db(resp.raw))
+        else:
+            raise typer.Exit(1)
+
+
+@app.command(short_help="Add a WebHook")
+def webhook(
+    name: str = typer.Argument(..., ),
+    urls: List[str] = typer.Argument(..., help="webhook urls",),
+    yes: bool = typer.Option(False, "-Y", help="Bypass confirmation prompts - Assume Yes"),
+    yes_: bool = typer.Option(False, "-y", hidden=True),
+    debug: bool = typer.Option(False, "--debug", envvar="ARUBACLI_DEBUG", help="Enable Additional Debug Logging",),
+    default: bool = typer.Option(False, "-d", is_flag=True, help="Use default central account",),
+    account: str = typer.Option("central_info",
+                                envvar="ARUBACLI_ACCOUNT",
+                                help="The Aruba Central Account to use (must be defined in the config)",),
+) -> None:
+    yes = yes_ if yes_ else yes
+
+    print("Adding WebHook: [cyan]{}[/cyan] with urls:\n  {}".format(name, '\n  '.join(urls)))
+    if yes or typer.confirm("\nProceed?", abort=True):
+        resp = cli.central.request(cli.central.add_webhook, name, urls)
+
+        cli.display_results(resp, tablefmt="action")
+        if not resp:
+            raise typer.Exit(1)
+
+
 @app.callback()
 def callback():
     """
