@@ -4,7 +4,7 @@
 import typer
 import sys
 import time
-from typing import Dict, List, Literal, Union
+from typing import Dict, List, Literal, Union, Tuple
 from pathlib import Path
 from rich.console import Console
 from rich import print
@@ -30,7 +30,7 @@ tty = utils.tty
 CASE_SENSITIVE_TOKENS = ["R", "U"]
 TableFormat = Literal["json", "yaml", "csv", "rich", "simple", "tabulate", "raw", "action"]
 MsgType = Literal["initial", "previous", "forgot", "will_forget", "previous_will_forget"]
-console = Console()
+console = Console(emoji=False)
 
 
 class CLICommon:
@@ -61,7 +61,7 @@ class CLICommon:
                 f"{typer.style(f'Using Account: {acct_clr}.', fg='magenta')}  "
                 f"{typer.style(f'Account setting is sticky.  ', fg='red', blink=True)}"
                 f"\n  {acct_clr} {typer.style(f'will be used for subsequent commands until', fg='magenta')}"
-                f"\n  {typer.style('--account <account name> or `-d` (revert to default). is used.', fg='magenta')}"
+                f"\n  {typer.style('--account <account name> or `-d` (revert to default). is used.', fg='magenta')}\n"
             )
 
         @property
@@ -70,28 +70,37 @@ class CLICommon:
                 f"{typer.style(f'Using previously specified account: ', fg='magenta')}"
                 f"{typer.style(self.account, fg='cyan', blink=True)}.  "
                 f"\n{typer.style('Use `--account <account name>` to switch to another account.', fg='magenta')}"
-                f"\n{typer.style('    or `-d` flag to revert to default account.', fg='magenta')}"
+                f"\n{typer.style('    or `-d` flag to revert to default account.', fg='magenta')}\n"
             )
 
         @property
         def forgot(self):
             return typer.style(
-                "Forget option set for account, and expiration has passed.  reverting to default account", fg="magenta"
+                "Forget option set for account, and expiration has passed.  reverting to default account\n", fg="magenta"
             )
 
         @property
         def will_forget(self):
-            return typer.style(
-                f"Forget options is configured, will revert to default account "
-                f'{typer.style(f"{config.forget_account_after} mins", fg="cyan")}'
-                f'{typer.style(" after last command.", fg="magenta")}',
-                fg="magenta",
-            )
+            will_forget_msg = f"[magenta]Forget options is configured, will revert to default account[/]\n"
+            will_forget_msg = f"{will_forget_msg}[cyan]{config.forget_account_after}[/][magenta] mins after last command[/]"
+            console = Console(emoji=False, record=True)
+            console.begin_capture()
+            console.print(will_forget_msg)
+            return console.end_capture()
+            # return typer.style(
+            #     f"Forget options is configured, will revert to default account "
+            #     f'{typer.style(f"{config.forget_account_after} mins", fg="cyan")}'
+            #     f'{typer.style(" after last command.", fg="magenta")}',
+            #     fg="magenta",
+            # )
 
         @property
         def previous_will_forget(self):
             return f"{self.previous}\n\n{self.will_forget}"
 
+    # TODO Can remove most of this and AcctMsg the bulk of messaging is now done in __init__
+    # as we need to know the account to load the correct cache.
+    # Initial --account still uses msg below in else
     def account_name_callback(self, ctx: typer.Context, account: str):
         if ctx.resilient_parsing:  # tab completion, return without validating
             return account
@@ -160,10 +169,24 @@ class CLICommon:
             return
 
         if default and config.sticky_account_file.is_file():
-            typer.secho("Using default central account", fg="bright_green")
+            typer.secho(" Using default central account", fg="bright_green")
             config.sticky_account_file.unlink()
             return default
 
+    @staticmethod
+    def send_cmds_node_callback(ctx: typer.Context, commands: Union[str, Tuple[str]]):
+        if ctx.resilient_parsing:  # tab completion, return without validating
+            return
+
+        # utils.json_print(ctx.__dict__)
+        # utils.json_print(ctx.parent.__dict__)
+        # utils.json_print(locals())
+        if ctx.params["kw1"].lower() == "all" and ctx.params["nodes"].lower() == "commands":
+            ctx.params["nodes"] = None
+            return tuple([ctx.params["kw2"], *commands])
+        else:
+            return commands
+
     @staticmethod
     def debug_callback(ctx: typer.Context, debug: bool):
         if ctx.resilient_parsing:  # tab completion, return without validating
@@ -187,6 +210,7 @@ class CLICommon:
     def normalize_tokens(token: str) -> str:
         return token.lower() if token not in CASE_SENSITIVE_TOKENS else token
 
+    # DEPRECATED
     def dev_completion(
         self,
         ctx: typer.Context,
@@ -233,8 +257,8 @@ class CLICommon:
                         "It looks like you are in the root of a git repo dir.\n"
                         "Exporting to out subdir."
                         )
-                config.outdir.mkdir(exist_ok=True)
-                outfile = config.outdir / outfile
+                    config.outdir.mkdir(exist_ok=True)
+                    outfile = config.outdir / outfile
 
             print(f"\n[cyan]Writing output to {outfile}... ", end="")
 
@@ -252,19 +276,22 @@ class CLICommon:
             if out_msg:
                 log.warning(out_msg, show=True)
 
+
     def _display_results(
         self,
         data: Union[List[dict], List[str], dict, None] = None,
         tablefmt: str = "rich",
         title: str = None,
         caption: str = None,
-        pager: bool = True,
+        pager: bool = False,
         outfile: Path = None,
         sort_by: str = None,
         reverse: bool = False,
+        stash: bool = True,
         pad: int = None,
         set_width_cols: dict = None,
         full_cols: Union[List[str], str] = [],
+        fold_cols: Union[List[str], str] = [],
         cleaner: callable = None,
         **cleaner_kwargs,
     ):
@@ -276,16 +303,23 @@ class CLICommon:
                 data = utils.listify(data)
 
             if sort_by and all(isinstance(d, dict) for d in data):
+                if sort_by not in data[0] and sort_by.replace("_", " ") in data[0]:
+                    sort_by = sort_by.replace("_", " ")
+
                 if not all([True if sort_by in d else False for d in data]):
-                    print(f"[dark_orange3]Warning: [cyan]{sort_by}[reset] does not appear to be a valid field")
+                    print(f":x: [dark_orange3]Error: [cyan]{sort_by}[reset] does not appear to be a valid field")
                     print("Valid Fields:\n----------\n{}\n----------".format("\n".join(data[0].keys())))
                 else:
                     try:
-                        data = sorted(data, key=lambda d: d[sort_by])
+                        type_ = str
+                        for d in data:
+                            if d[sort_by] is not None:
+                                type_ = type(d[sort_by])
+                                break
+                        data = sorted(data, key=lambda d: d[sort_by] if d[sort_by] != "-" else 0 or 0 if type_ == int else "")
                     except TypeError as e:
                         print(
-                            f"[dark_orange3]Warning:[reset] Unable to sort by [cyan]{sort_by}. "
-                            f"[dark_orange3]TypeError:[reset] {e}"
+                            f":x: [dark_orange3]Warning:[reset] Unable to sort by [cyan]{sort_by}.\n   {e.__class__.__name__}: {e} "
                         )
 
             if reverse:
@@ -294,20 +328,37 @@ class CLICommon:
             if self.raw_out and tablefmt in ["simple", "rich"]:
                 tablefmt = "json"
 
-            outdata = utils.output(
-                data,
-                tablefmt,
-                title=title,
-                caption=caption,
-                account=None if config.account in ["central_info", "account"] else config.account,
-                config=config,
-                set_width_cols=set_width_cols,
-                full_cols=full_cols,
-            )
+            # TODO make sure "account" is not valid then remove from list below
+            if config.account == "account":
+                log.warning("DEV NOTE account is 'account'", show=True)
+
+            kwargs = {
+                "outdata": data,
+                "tablefmt": tablefmt,
+                "title": title,
+                "caption": caption,
+                "account": None if config.account in ["central_info", "default", "account"] else config.account,
+                "config": config,
+                "set_width_cols": set_width_cols,
+                "full_cols": full_cols,
+                "fold_cols": fold_cols,
+            }
+            outdata = utils.output(**kwargs)
+
+            if stash:
+                config.last_command_file.write_text(
+                    json.dumps({k: v for k, v in kwargs.items() if k != "config"})
+                )
+
             typer.echo_via_pager(outdata) if pager and tty and len(outdata) > tty.rows else typer.echo(outdata)
 
+            if "Limit:" not in outdata and caption is not None and cleaner and cleaner.__name__ != "parse_caas_response":
+                print(caption)
+
             if outfile and outdata:
                 self.write_file(outfile, outdata.file)
+        else:
+            log.warning(f"No data passed to _display_output {title} {caption}")
 
     def display_results(
         self,
@@ -316,15 +367,17 @@ class CLICommon:
         tablefmt: TableFormat = "rich",
         title: str = None,
         caption: str = None,
-        pager: bool = True,
+        pager: bool = False,
         outfile: Path = None,
         sort_by: str = None,
-        reverse: bool = None,
+        reverse: bool = False,
+        stash: bool = True,
         pad: int = None,
         exit_on_fail: bool = False,
         ok_status: Union[int, List[int], Dict[int, str]] = None,
         set_width_cols: dict = None,
         full_cols: Union[List[str], str] = [],
+        fold_cols: Union[List[str], str] = [],
         cleaner: callable = None,
         **cleaner_kwargs,
     ) -> None:
@@ -347,6 +400,8 @@ class CLICommon:
             outfile (Path, optional): path/file of output file. Defaults to None.
             sort_by (Union[str, List[str], None] optional): column or columns to sort output on.
             reverse (bool, optional): reverse the output.
+            stash (bool, optional): stash (cache) the output of the command.  The CLI can re-display with
+                show last.  Default: True
             ok_status (Union[int, List[int], Tuple[int, str], List[Tuple[int, str]]], optional): By default
                 responses with status_code 2xx are considered OK and are rendered as green by
                 Output class.  provide int or list of int to override additional status_codes that
@@ -362,30 +417,33 @@ class CLICommon:
         if pad:
             log.error("Deprecated pad parameter referenced in display_results", show=True)
 
-        pager = False if config.no_pager else pager
-
         if resp is not None:
             resp = utils.listify(resp)
 
             # update caption with rate limit
             if resp[-1].rl:
-                rl_str = f"[italic dark_olive_green2]{resp[-1].rl}".lstrip()
-                caption = f"{caption} {rl_str}" if caption else rl_str
+                rl_str = f"[italic dark_olive_green2]{resp[-1].rl}[/]".lstrip()
+                caption = f"{caption}\n  {rl_str}" if caption else f"  {rl_str}"
 
             for idx, r in enumerate(resp):
                 # Multi request url line
+                m_colors = {
+                    "DELETE": "red",
+                    "GET": "bright_green",
+                    "PATH": "dark_orange3",
+                }
+                fg = "bright_green" if r else "red"
                 if len(resp) > 1:
-                    _url = r.url if not hasattr(r.url, "path") else r.url.path
-                    # typer.secho(f"Request {idx + 1} [{r.method}: {_url}] Response:", fg="cyan")
+                    _url = r.url if not hasattr(r.url, "raw_path_qs") else r.url.path
+                    m_color = m_colors.get(r.method, "reset")
                     print(
-                        f"Request [bright_green]{idx + 1}[/bright_green]. [[bright_green]{r.method}[/bright_green]: "
-                        f"[bright_green]{_url}[/bright_green]] Response:"
+                        f"Request {idx + 1} [[{m_color}]{r.method}[reset]: "
+                        f"[cyan]{_url}[/cyan]]\n [fg]Response[reset]:"
                     )
                 if self.raw_out:
                     tablefmt = "raw"
 
                 if not r or tablefmt in ["action", "raw"]:
-                    fg = "green" if r else "red"
 
                     if tablefmt == "raw":
                         # dots = f"[{fg}]{'.' * 16}[/{fg}]"
@@ -402,14 +460,11 @@ class CLICommon:
                             self.write_file(outfile, r.raw)
 
                     else:
-                        # typer.secho(str(r), fg=fg)
                         print(f"[{fg}]{r}")
 
                     if idx + 1 == len(resp):
                         console.print(f"\n{rl_str}")
 
-                    if not r and exit_on_fail:
-                        raise typer.Exit(1)
                 else:
                     self._display_results(
                         r.output,
@@ -420,13 +475,22 @@ class CLICommon:
                         outfile=outfile,
                         sort_by=sort_by,
                         reverse=reverse,
+                        stash=stash,
                         pad=pad,
                         set_width_cols=set_width_cols,
                         full_cols=full_cols,
+                        fold_cols=fold_cols,
                         cleaner=cleaner,
                         **cleaner_kwargs
                     )
 
+            # TODO make elegant caas send-cmds uses this logic
+            if cleaner and cleaner.__name__ == "parse_caas_response":
+                print(caption)
+
+            if exit_on_fail and not all([r.ok for r in resp]):
+                raise typer.Exit(1)
+
         elif data:
             self._display_results(
                 data,
@@ -437,9 +501,11 @@ class CLICommon:
                 outfile=outfile,
                 sort_by=sort_by,
                 reverse=reverse,
+                stash=stash,
                 pad=pad,
                 set_width_cols=set_width_cols,
                 full_cols=full_cols,
+                fold_cols=fold_cols,
                 cleaner=cleaner,
                 **cleaner_kwargs
             )
