@@ -31,6 +31,10 @@ class GatewayRole(str, Enum):
     wlan = "wlan"
 
 
+class StartArgs(str, Enum):
+    hook_proxy = "hook-proxy"
+
+
 # wrapping keys from return for some calls that have no value
 STRIP_KEYS = [
     "data",
@@ -55,6 +59,8 @@ STRIP_KEYS = [
     "neighbor_aps",
     "events",
     "notifications",
+    "settings",
+    "items",
 ]
 
 
@@ -240,11 +246,13 @@ class ArgToWhat:
         self.wlans = self.wlan = self.ssids = self.ssid = "wlans"
         self.run = self.running = "run"
         self.routes = self.route = "routes"
+        self.webhooks = self.webhook = "webhooks"
 
     def _init_update(self):
         self.template = self.templates = "template"
         self.variables = self.variable = "variables"
         self.group = self.groups = "group"
+        self.webhooks = self.webhook = "webhook"
 
     def _init_rename(self):
         self.group = self.groups = "group"
@@ -255,6 +263,7 @@ class ArgToWhat:
         self.group = self.groups = "group"
         self.certificate = self.certs = self.certificates = self.cert = "certificate"
         self.wlan = self.wlans = "wlan"
+        self.webhooks = self.webhook = "webhook"
 
     def _init_upgrade(self):
         self.device = self.devices = self.dev = "device"
@@ -262,9 +271,14 @@ class ArgToWhat:
         self.swarm = "swarm"
 
     def _init_add(self):
+        self.site = self.sites = "site"
         self.group = self.groups = "group"
         self.wlan = self.wlans = "wlan"
         self.device = self.devices = self.dev = "device"
+        self.webhooks = self.webhook = "webhook"
+
+    def _init_test(self):
+        self.webhooks = self.webhook = "webhook"
 
     def _init_clone(self):
         self.group = self.groups = "group"
@@ -272,6 +286,9 @@ class ArgToWhat:
     def _init_bounce(self):
         self.interface = self.port = self.ports = self.interfaces = "interface"
 
+    def _init_caas(self):
+        self.send_cmd = self.send_cmds = "send_cmds"
+
     def __call__(self, key: Union[ShowArgs, str], default: str = None, cmd: str = "show") -> str:
         if cmd != "show":
             if hasattr(self, f"_init_{cmd}"):
@@ -317,12 +334,12 @@ class LibToAPI:
 
         # from CentralApi consistent value to Random API value
         self.monitoring_to_api = {
-             "gw": "gateways",
-             "ap": "aps",
-             "switch": "switches",
-             "cx": "switches",
-             "sw": "Switches"
-         }
+            "gw": "gateways",
+            "ap": "aps",
+            "switch": "switches",
+            "cx": "switches",
+            "sw": "switches"
+        }
         self.site_to_api = {
             "gw": "CONTROLLER",
             "ap": "IAP",
@@ -441,6 +458,12 @@ class ShowArg2(str, Enum):
     certs = "certs"
 
 
+class ShowHookProxyArgs(str, Enum):
+    logs = "logs"
+    pid = "pid"
+    port = "port"
+
+
 # TODO remove once all refs removed
 class SortOptions(str, Enum):
     name_asc = "+name"
@@ -483,6 +506,29 @@ class SortCertOptions(str, Enum):
     sha1_checksum = "sha1_checksum"
 
 
+class SendCmdArgs(str, Enum):
+    device = "device"
+    site = "site"
+    group = "group"
+    all = "all"
+    file = "file"
+
+
+class SendCmdArg2(str, Enum):
+    commands = "commands"
+
+
+class SortSiteOptions(str, Enum):
+    name = "name"
+    id = "id"
+    address = "address"
+    city = "city"
+    state = "state"
+    zipcode = "zipcode"
+    country = "country"
+    associated_devices = "associated_devices"
+
+
 class SortVlanOptions(str, Enum):
     name = "name"
     pvid = "pvid"
@@ -546,11 +592,13 @@ class IdenMetaVars:
         self.dev = "[name|ip|mac|serial]"
         self.dev_many = "[name|ip|mac|serial] ... (multiple allowed)"
         self.group_many = "[GROUP NAME] ... (multiple allowed)"
-        self.site = "[name|site_id|address|city|state|zip]"
+        self.site = "[name|site id|address|city|state|zip]"
         self.client = "[username|ip|mac]"
         self.dev_words = f"Optional Identifying Attribute: {self.dev}"
         self.generic_dev_types = "[ap|gw|switch]"
         self.dev_types = "[ap|gw|cx|sw]"
+        self.group_or_dev = f"device {self.dev.upper()} | group [GROUP]"
+        self.group_or_dev_or_site = "[DEVICE|\"all\"|GROUP|SITE]"
 
 
 class LogSortBy(str, Enum):
@@ -563,7 +611,6 @@ class LogSortBy(str, Enum):
     ip = "ip"
     user = "user"
     id = "id"
-    has_details = "has_details"
 
 
 LIB_DEV_TYPE = {
@@ -586,70 +633,130 @@ def get_cencli_devtype(dev_type: str) -> str:
 
 
 state_abbrev_to_pretty = {
-  "AL": "Alabama",
-  "AK": "Alaska",
-  "AZ": "Arizona",
-  "AR": "Arkansas",
-  "CA": "California",
-  "CO": "Colorado",
-  "CT": "Connecticut",
-  "DE": "Delaware",
-  "FL": "Florida",
-  "GA": "Georgia",
-  "HI": "Hawaii",
-  "ID": "Idaho",
-  "IL": "Illinois",
-  "IN": "Indiana",
-  "IA": "Iowa",
-  "KS": "Kansas",
-  "KY": "Kentucky",
-  "LA": "Louisiana",
-  "ME": "Maine",
-  "MD": "Maryland",
-  "MA": "Massachusetts",
-  "MI": "Michigan",
-  "MN": "Minnesota",
-  "MS": "Mississippi",
-  "MO": "Missouri",
-  "MT": "Montana",
-  "NE": "Nebraska",
-  "NV": "Nevada",
-  "NH": "New Hampshire",
-  "NJ": "New Jersey",
-  "NM": "New Mexico",
-  "NY": "New York",
-  "NC": "North Carolina",
-  "ND": "North Dakota",
-  "OH": "Ohio",
-  "OK": "Oklahoma",
-  "OR": "Oregon",
-  "PA": "Pennsylvania",
-  "RI": "Rhode Island",
-  "SC": "South Carolina",
-  "SD": "South Dakota",
-  "TN": "Tennessee",
-  "TX": "Texas",
-  "UT": "Utah",
-  "VT": "Vermont",
-  "VA": "Virginia",
-  "WA": "Washington",
-  "WV": "West Virginia",
-  "WI": "Wisconsin",
-  "WY": "Wyoming",
-  "DC": "District of Columbia",
-  "AS": "American Samoa",
-  "GU": "Guam",
-  "MP": "Northern Mariana Islands",
-  "PR": "Puerto Rico",
-  "UM": "United States Minor Outlying Islands",
-  "VI": "U.S. Virgin Islands"
+    "AL": "Alabama",
+    "AK": "Alaska",
+    "AZ": "Arizona",
+    "AR": "Arkansas",
+    "CA": "California",
+    "CO": "Colorado",
+    "CT": "Connecticut",
+    "DE": "Delaware",
+    "FL": "Florida",
+    "GA": "Georgia",
+    "HI": "Hawaii",
+    "ID": "Idaho",
+    "IL": "Illinois",
+    "IN": "Indiana",
+    "IA": "Iowa",
+    "KS": "Kansas",
+    "KY": "Kentucky",
+    "LA": "Louisiana",
+    "ME": "Maine",
+    "MD": "Maryland",
+    "MA": "Massachusetts",
+    "MI": "Michigan",
+    "MN": "Minnesota",
+    "MS": "Mississippi",
+    "MO": "Missouri",
+    "MT": "Montana",
+    "NE": "Nebraska",
+    "NV": "Nevada",
+    "NH": "New Hampshire",
+    "NJ": "New Jersey",
+    "NM": "New Mexico",
+    "NY": "New York",
+    "NC": "North Carolina",
+    "ND": "North Dakota",
+    "OH": "Ohio",
+    "OK": "Oklahoma",
+    "OR": "Oregon",
+    "PA": "Pennsylvania",
+    "RI": "Rhode Island",
+    "SC": "South Carolina",
+    "SD": "South Dakota",
+    "TN": "Tennessee",
+    "TX": "Texas",
+    "UT": "Utah",
+    "VT": "Vermont",
+    "VA": "Virginia",
+    "WA": "Washington",
+    "WV": "West Virginia",
+    "WI": "Wisconsin",
+    "WY": "Wyoming",
+    "DC": "District of Columbia",
+    "AS": "American Samoa",
+    "GU": "Guam",
+    "MP": "Northern Mariana Islands",
+    "PR": "Puerto Rico",
+    "UM": "United States Minor Outlying Islands",
+    "VI": "U.S. Virgin Islands"
 }
 
 # invert
 state_pretty_to_abbrev = dict(map(reversed, state_abbrev_to_pretty.items()))
 
+class SiteStates(str, Enum):
+    AL = Alabama = "Alabama",
+    AK = Alaska = "Alaska",
+    AZ = Arizona = "Arizona",
+    AR = Arkansas = "Arkansas",
+    CA = California = "California",
+    CO = Colorado = "Colorado",
+    CT = Connecticut = "Connecticut",
+    DE = Delaware = "Delaware",
+    FL = Florida = "Florida",
+    GA = Georgia = "Georgia",
+    HI = Hawaii = "Hawaii",
+    ID = Idaho = "Idaho",
+    IL = Illinois = "Illinois",
+    IN = Indiana = "Indiana",
+    IA = Iowa = "Iowa",
+    KS = Kansas = "Kansas",
+    KY = Kentucky = "Kentucky",
+    LA = Louisiana = "Louisiana",
+    ME = Maine = "Maine",
+    MD = Maryland = "Maryland",
+    MA = Massachusetts = "Massachusetts",
+    MI = Michigan = "Michigan",
+    MN = Minnesota = "Minnesota",
+    MS = Mississippi = "Mississippi",
+    MO = Missouri = "Missouri",
+    MT = Montana = "Montana",
+    NE = Nebraska = "Nebraska",
+    NV = Nevada = "Nevada",
+    NH = New_Hampshire = "New Hampshire",
+    NJ = New_Jersey = "New Jersey",
+    NM = New_Mexico = "New Mexico",
+    NY = New_York = "New York",
+    NC = North_Carolina = "North Carolina",
+    ND = North_Dakota = "North Dakota",
+    OH = Ohio = "Ohio",
+    OK = Oklahoma = "Oklahoma",
+    OR = Oregon = "Oregon",
+    PA = Pennsylvania = "Pennsylvania",
+    RI = Rhode_Island = "Rhode Island",
+    SC = South_Carolina = "South Carolina",
+    SD = South_Dakota = "South Dakota",
+    TN = Tennessee = "Tennessee",
+    TX = Texas = "Texas",
+    UT = Utah = "Utah",
+    VT = Vermont = "Vermont",
+    VA = Virginia = "Virginia",
+    WA = Washington = "Washington",
+    WV = West_Virginia = "West Virginia",
+    WI = Wisconsin = "Wisconsin",
+    WY = Wyoming = "Wyoming",
+    DC = District_of_Columbia = "District of Columbia",
+    AS = American_Samoa = "American Samoa",
+    GU = Guam = "Guam",
+    MP = Northern_Mariana_Islands = "Northern Mariana Islands",
+    PR = Puerto_Rico = "Puerto Rico",
+    UM = United_States_Minor_Outlying_Islands = "United States Minor Outlying Islands",
+    VI = US_Virgin_Islands = "U.S. Virgin Islands"
+
 NO_LOAD_COMMANDS = [
     "show config cencli",
+    "show last"
 ]
 NO_LOAD_FLAGS = [
     "--help",
