@@ -1,6 +1,7 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
+import asyncio
 import json
 import os
 import shutil
@@ -75,15 +76,17 @@ class Utils:
         self.Mac = Mac
 
     def user_input_bool(self, question):
+
         """Ask User Y/N Question require Y/N answer
 
         Error and re-prompt if user's response is not valid
         Appends '? (y/n): ' to question/prompt provided
 
         Params:
-            question:str, The Question to ask
+            question (str): The Question to ask
+
         Returns:
-            answer:bool, Users Response yes=True
+            answer (bool): Users Response yes=True
         """
         valid_answer = ["yes", "y", "no", "n"]
         try:
@@ -234,13 +237,13 @@ class Utils:
 
         return r
 
-    # TODO depricated validate not used, moved to Response
+    # TODO deprecated validate not used, moved to Response
     @staticmethod
     def get_multiline_input(prompt: str = None, print_func: callable = print,
                             return_type: str = None, abort_str: str = "exit", **kwargs) -> Union[List[str], dict, str]:
         def _get_multiline_sub(prompt: str = prompt, print_func: callable = print_func, **kwargs):
             prompt = prompt or \
-                "Enter/Paste your content. Then Ctrl-D or Ctrl-Z ( windows ) to submit.\n Enter 'exit' to abort"
+                "Enter/Paste your content. Then Ctrl-D or Ctrl-Z -> Enter ( windows ) to submit.\n Enter 'exit' to abort"
             print_func(prompt, **kwargs)
             contents, line = [], ''
             while line.strip().lower() != abort_str:
@@ -274,11 +277,11 @@ class Utils:
         return contents
 
     @staticmethod
-    def strip_none(_dict: Union[dict, None], strip_empty_obj: bool = False) -> Any:
-        """strip all keys from a dict where value is NoneType
+    def strip_none(data: Union[dict, list, None], strip_empty_obj: bool = False) -> Any:
+        """strip all keys from a list or dict where value is NoneType
 
         args:
-            _dict: The dictionary object to have all keys removed that have None as value
+            data (Union[dict, list, None]): The iterable object to have all items/keys where value is None
             strip_empty_obj (bool): If True will strip keys that have empty objects as
                 well as NoneType for value.
 
@@ -286,13 +289,18 @@ class Utils:
             (Any) Return same type that is provided to method as first argument.
             If dict is provided it returns modified dict.
         """
-        if not isinstance(_dict, dict):
-            return _dict
-
-        if not strip_empty_obj:
-            return {k: v if not callable(v) else v.__name__ for k, v in _dict.items() if v is not None}
+        if isinstance(data, dict):
+            if not strip_empty_obj:
+                return {k: v if not callable(v) else v.__name__ for k, v in data.items() if v is not None}
+            else:
+                return {k: v for k, v in data.items() if not isinstance(v, bool) and v}
+        elif isinstance(data, (list, tuple)):
+            if strip_empty_obj:
+                return type(data)(d for d in data if d)
+            return type(data)(d for d in data if d is not None)
         else:
-            return {k: v for k, v in _dict.items() if not isinstance(v, bool) and v}
+            return data
+
 
     class Output:
         def __init__(self, rawdata: str = "", prettydata: str = "", config=None):
@@ -321,6 +329,9 @@ class Utils:
             for line in out:
                 yield line
 
+        def __contains__(self, item):
+            return item in self.file
+
         def sanitize_strings(self, strings: str, config=None) -> str:
             config = config or self.config
             if config and config.sanitize and config.sanatize_file.is_file():
@@ -382,6 +393,7 @@ class Utils:
         config=None,
         set_width_cols: dict = None,
         full_cols: Union[List[str], str] = [],
+        fold_cols: Union[List[str], str] = [],
         ok_status: Union[int, List[int], Tuple[int, str], List[Tuple[int, str]]] = None,
     ) -> str:
         # log.debugv(f"data passed to output():\n{pprint(outdata, indent=4)}")
@@ -416,9 +428,9 @@ class Utils:
                                                 box=SIMPLE
                                                 )
                             _ = [inner_table.add_row(*[self.do_pretty(kk, str(vv)) for kk, vv in v.items()]) for v in val]
-                            with console.capture():
-                                console.print(inner_table)
-                            inner_dict[key] = console.export_text()
+                            console.begin_capture()
+                            console.print(inner_table)
+                            inner_dict[key] = console.end_capture()
                         elif val and tablefmt == "tabulate" and hasattr(val[0], 'keys'):
                             inner_table = tabulate(val, headers="keys", tablefmt=tablefmt)
                             inner_dict[key] = inner_table
@@ -475,7 +487,7 @@ class Utils:
             from rich.text import Text
             # from rich.progress import Progress
             from centralcli import constants
-            console = Console(record=True)
+            console = Console(record=True, emoji=False)
 
             customer_id, customer_name = "", ""
             # outdata = self.listify(outdata)
@@ -496,7 +508,7 @@ class Utils:
                     row_styles=['none', 'dark_sea_green']
                 )
 
-                fold_cols = ['description']
+                fold_cols = [*fold_cols, 'description']
                 _min_max = {'min': 10, 'max': 30}
                 set_width_cols = set_width_cols or {'name': _min_max, 'model': _min_max}
                 # default full cols #TODO clean this up
@@ -530,14 +542,15 @@ class Utils:
 
                 data_header = f"--\n{'Customer ID:':15}{customer_id}\n{'Customer Name:':15} {customer_name}\n--\n"
 
-                with console.capture():
-                    console.print(table)
+                # TODO look into this. console.capture stopped working reliably this works
+                console.begin_capture()
+                console.print(table)
+                table_data = console.end_capture()
+                raw_data = typer.unstyle(table_data)
 
-                raw_data = console.export_text(clear=False)
-                table_data = console.export_text(styles=True)
-
-                raw_data = f"{data_header}{raw_data}" if customer_id else f"{raw_data}"
-                table_data = f"{data_header}{table_data}" if customer_id else f"{table_data}"
+                if customer_id:
+                    raw_data = f"{data_header}{raw_data}"
+                    table_data = f"{data_header}{table_data}"
 
         elif tablefmt == "tabulate":
             customer_id = customer_name = ""
@@ -624,3 +637,7 @@ class Utils:
             return ", ".join(text)
         else:
             raise TypeError(f"{type(text)}: text attribute should be str, bool, or list of str.")
+
+    @staticmethod
+    def chunker(seq, size):
+        return [seq[pos:pos + size] for pos in range(0, len(seq), size)]
